@@ -1,13 +1,13 @@
 ;;; jinx.el --- Enchanted Spell Checker -*- lexical-binding: t -*-
 
-;; Copyright (C) 2023-2024 Free Software Foundation, Inc.
+;; Copyright (C) 2023-2025 Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2023
-;; Version: 1.9
-;; Package-Requires: ((emacs "27.1") (compat "29.1.4.4"))
-;; Homepage: https://github.com/minad/jinx
+;; Version: 1.12
+;; Package-Requires: ((emacs "28.1") (compat "30"))
+;; URL: https://github.com/minad/jinx
 ;; Keywords: convenience, text
 
 ;; This file is part of GNU Emacs.
@@ -41,11 +41,11 @@
 ;; also safely co-exist with Emacs's built-in spell-checker Ispell.
 
 ;; Jinx's high performance and low resource usage comes from directly
-;; calling the widely-used API of the Enchant library (see
-;; https://abiword.github.io/enchant/).  Jinx automatically compiles
+;; calling the API of the Enchant library, see
+;; https://rrthomas.github.io/enchant/.  Jinx automatically compiles
 ;; jinx-mod.c and loads the dynamic module at startup.  By binding
-;; directly to the native Enchant API, Jinx avoids the slower backend
-;; process communication with Aspell.
+;; directly to the native Enchant API, Jinx avoids slower
+;; inter-process communication.
 
 ;; See the manual for further information.
 
@@ -61,7 +61,8 @@
 (defgroup jinx nil
   "Enchanted Spell Checker."
   :link '(info-link :tag "Info Manual" "(jinx)")
-  :link '(url-link :tag "Homepage" "https://github.com/minad/jinx")
+  :link '(url-link :tag "Website" "https://github.com/minad/jinx")
+  :link '(url-link :tag "Wiki" "https://github.com/minad/jinx/wiki")
   :link '(emacs-library-link :tag "Library Source" "jinx.el")
   :group 'text
   :prefix "jinx-")
@@ -108,8 +109,8 @@ checking."
 (defcustom jinx-camel-modes
   '(java-mode java-ts-mode js-mode js-ts-mode ruby-mode ruby-ts-mode rust-mode
     rust-ts-mode haskell-mode kotlin-mode swift-mode csharp-mode csharp-ts-mode
-    objc-mode typescript-ts-mode typescript-mode python-mode python-ts-mode
-    dart-mode go-mode go-ts-mode scala-mode groovy-mode)
+    objc-mode typescript-ts-mode typescript-mode tsx-ts-mode python-mode
+    python-ts-mode dart-mode go-mode go-ts-mode scala-mode groovy-mode)
   "Modes where camelCase or PascalCase words should be accepted.
 Set to t to enable camelCase everywhere."
   :type '(choice (const t) (repeat symbol)))
@@ -139,7 +140,7 @@ Set to t to enable camelCase everywhere."
      rst-literal rst-external rst-directive rst-definition rst-reference)
     (sgml-mode
      font-lock-function-name-face font-lock-variable-name-face)
-    (emacs-lisp-mode
+    (lisp-data-mode
      font-lock-constant-face font-lock-warning-face)
     (message-mode
      message-separator message-header-cc message-header-name
@@ -262,6 +263,10 @@ checking."
 
 ;;;; Internal variables
 
+(defvar jinx--compile-flags
+  '("-I." "-O2" "-Wall" "-Wextra" "-fPIC" "-shared")
+  "List of compile flags passed to the C compiler.")
+
 (defvar jinx--reschedule-hooks
   '(window-selection-change-functions window-scroll-functions
     window-state-change-hook post-command-hook)
@@ -291,7 +296,13 @@ Predicate may return a position to skip forward.")
     (modify-syntax-entry '(#x1cf00 . #x1d7ff) "_" st)  ;; Znamenny Musical - Math. Alpha.
     (modify-syntax-entry '(#x1ee00 . #x1fbff) "_" st)  ;; Arabic Math. - Legacy Computing
     st)
-  "Base syntax table for spell checking.")
+  "Parent syntax table of `jinx--syntax-table'.")
+
+(defvar jinx--syntax-overrides
+  '((?' . "w")
+    (?’ . "w")
+    (?. . "."))
+  "Syntax table overrides used for `jinx--syntax-table'.")
 
 (defvar jinx--select-keys
   "123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
@@ -318,8 +329,14 @@ Predicate may return a position to skip forward.")
 (defvar-local jinx--dicts nil
   "List of dictionaries.")
 
+(defvar jinx--dicts-hash (make-hash-table :test #'equal :weakness 'value)
+  "Weak hash table of all loaded dictionaries.")
+
 (defvar-local jinx--syntax-table nil
-  "Syntax table used during checking.")
+  "Syntax table used during checking.
+The table inherits from `jinx--base-syntax-table'.  The table is
+configured according to the word characters defined by the local
+dictionaries.  Afterwards `jinx--syntax-overrides' are applied.")
 
 (defvar-local jinx--session-words nil
   "List of words accepted in this session.")
@@ -370,19 +387,21 @@ Predicate may return a position to skip forward.")
               (cl-loop for f in face thereis (memq f jinx--exclude-faces))
             (memq face jinx--exclude-faces))))))
 
-(defun jinx--word-valid-p (start)
-  "Return non-nil if word at START is valid."
-  (let ((word (buffer-substring-no-properties start (point)))
-        case-fold-search)
-    (or (member word jinx--session-words)
-        ;; Allow capitalized words
+(defun jinx--word-valid-p (word)
+  "Return non-nil if WORD is valid.
+WORD can also be a start position."
+  (unless (stringp word)
+    (setq word (buffer-substring-no-properties word (point))))
+  (or (member word jinx--session-words)
+      ;; Allow capitalized words
+      (let (case-fold-search)
         (and (string-match-p "\\`[[:upper:]][[:lower:]]+\\'" word)
              (cl-loop
               for w in jinx--session-words
               thereis (and (string-equal-ignore-case word w)
-                           (string-match-p "\\`[[:lower:]]+\\'" w))))
-        (cl-loop for dict in jinx--dicts
-                 thereis (jinx--mod-check dict word)))))
+                           (string-match-p "\\`[[:lower:]]+\\'" w)))))
+      (cl-loop for dict in jinx--dicts
+               thereis (jinx--mod-check dict word))))
 
 ;;;; Internal functions
 
@@ -396,16 +415,11 @@ Predicate may return a position to skip forward.")
 This function is a modification hook for the overlay."
   (delete-overlay overlay))
 
-(defun jinx--find-visible-pending (start end flag)
-  "Find (in)visible and (non-)pending region between START and END.
-FLAG must be t or nil."
-  (while (and (< start end)
-              (eq flag
-                  (not (and (get-text-property start 'jinx--pending)
-                            (not (invisible-p start))))))
-    (let ((next (next-single-char-property-change
-                 start 'jinx--pending nil
-                 (next-single-char-property-change start 'invisible nil end))))
+(defun jinx--find-visible (start end visible)
+  "Find (in)visible region between START and END.
+VISIBLE must be nil or t."
+  (while (and (< start end) (eq (not visible) (not (invisible-p start))))
+    (let ((next (next-single-char-property-change start 'invisible nil end)))
       ;; END can be outside the buffer if the buffer size has changed in
       ;; between. Then `next-single-property-change' will return (point-max)
       ;; instead of END. See gh:minad/jinx#156.
@@ -414,18 +428,21 @@ FLAG must be t or nil."
 
 (defun jinx--check-pending (start end)
   "Check pending visible region between START and END."
-  (let ((pos start)
-        (retry (and (eq (window-buffer) (current-buffer))
+  (let ((retry (and (eq (window-buffer) (current-buffer))
                     (symbolp real-last-command)
                     (string-match-p "self-insert-command\\'"
                                     (symbol-name real-last-command))
                     (window-point))))
-    (while (< pos end)
-      (let* ((from (jinx--find-visible-pending pos end t))
-             (to (jinx--find-visible-pending from end nil)))
-        (if (< from to)
-            (setq pos (jinx--check-region from to retry))
-          (setq pos to))))))
+    (while (< start end)
+      (let* ((vfrom (jinx--find-visible start end t))
+             (vto (jinx--find-visible vfrom end nil)))
+        (while (< vfrom vto)
+          (let* ((pfrom (or (text-property-any vfrom vto 'jinx--pending t) vto))
+                 (pto (or (text-property-not-all pfrom vto 'jinx--pending t) vto)))
+            (when (< pfrom pto)
+              (jinx--check-region pfrom pto retry))
+            (setq vfrom pto)))
+        (setq start vto)))))
 
 (defun jinx--check-region (start end &optional retry)
   "Check region between START and END.
@@ -480,8 +497,7 @@ position."
               (remove-list-of-text-properties start end '(jinx--pending))
               (when retry-start
                 (put-text-property retry-start retry-end 'jinx--pending t)))))
-      (set-syntax-table st)))
-  end)
+      (set-syntax-table st))))
 
 (defun jinx--get-overlays (start end &optional visible)
   "Return misspelled word overlays between START and END.
@@ -504,11 +520,10 @@ If VISIBLE is non-nil, only include visible overlays.
 If CHECK is non-nil, always check first."
   (or (and (not check) (jinx--get-overlays start end visible))
       (progn
-        (let (set-message-function) ;; bug#63253: Broken `with-delayed-message'
-          (with-delayed-message (1 "Fontifying...")
-            (jinx--in-base-buffer #'jit-lock-fontify-now start end))
-          (with-delayed-message (1 "Checking...")
-            (jinx--check-region start end)))
+        (with-delayed-message (1 "Fontifying...")
+          (jinx--in-base-buffer #'jit-lock-fontify-now start end))
+        (with-delayed-message (1 "Checking...")
+          (jinx--check-region start end))
         (jinx--get-overlays start end visible))
       (user-error (if visible "No misspelled word in visible text"
                     "No misspelled word in whole buffer"))))
@@ -592,12 +607,11 @@ If CHECK is non-nil, always check first."
                                    (or (locate-library c-name t)
                                        (error "Jinx: %s not found" c-name))))
                (command
-                `(,cc "-I." "-O2" "-Wall" "-Wextra" "-fPIC" "-shared"
-                  "-o" ,mod-name ,c-name
+                `(,cc ,@jinx--compile-flags "-o" ,mod-name ,c-name
                   ,@(split-string-and-unquote
                      (condition-case nil
                          (car (process-lines "pkg-config" "--cflags" "--libs" "enchant-2"))
-                       (error "-I/usr/include/enchant-2 -lenchant-2"))))))
+                       (error "-I/usr/include/enchant-2 -I/usr/local/include/enchant-2 -L/usr/local/lib -lenchant-2"))))))
           (with-current-buffer (get-buffer-create "*jinx module compilation*")
             (let ((inhibit-read-only t))
               (erase-buffer)
@@ -627,42 +641,25 @@ If CHECK is non-nil, always check first."
 (defun jinx--invisible-open-temporarily ()
   "Temporarily open overlays which hide the current line.
 See `isearch-open-necessary-overlays' and `isearch-open-overlay-temporary'."
-  (if (and (derived-mode-p 'org-mode)
-           (fboundp 'org-fold-show-set-visibility)
-           (fboundp 'org-fold-core-get-regions)
-           (fboundp 'org-fold-core-region))
-      ;; New Org 9.6 fold-core API
-      (let ((regions (delq nil (org-fold-core-get-regions
-                                :with-markers t :from (point-min) :to (point-max)))))
-        (let ((inhibit-redisplay t)) ;; HACK: Prevent flicker due to premature redisplay
-          (org-fold-show-set-visibility 'canonical))
-        (list (lambda ()
-                (cl-loop for (beg end spec) in regions do
-                         (org-fold-core-region beg end t spec)))))
-    (let (restore)
-      (dolist (ov (overlays-in (pos-bol) (pos-eol)))
-        (let ((inv (overlay-get ov 'invisible)))
-          (when (and (invisible-p inv) (overlay-get ov 'isearch-open-invisible))
-            (push (if-let ((fun (overlay-get ov 'isearch-open-invisible-temporary)))
-                      (progn
-                        (funcall fun ov nil)
-                        (lambda () (funcall fun ov t)))
-                    (overlay-put ov 'invisible nil)
-                    (lambda () (overlay-put ov 'invisible inv)))
-                  restore))))
-      restore)))
+  (let (restore)
+    (dolist (ov (overlays-in (pos-bol) (pos-eol)) restore)
+      (let ((inv (overlay-get ov 'invisible)))
+        (when (and (invisible-p inv) (overlay-get ov 'isearch-open-invisible))
+          (push (if-let ((fun (overlay-get ov 'isearch-open-invisible-temporary)))
+                    (progn
+                      (funcall fun ov nil)
+                      (lambda () (funcall fun ov t)))
+                  (overlay-put ov 'invisible nil)
+                  (lambda () (overlay-put ov 'invisible inv)))
+                restore))))))
 
 (defun jinx--invisible-open-permanently ()
   "Open overlays which hide the current line.
 See `isearch-open-necessary-overlays' and `isearch-open-overlay-temporary'."
-  (if (and (derived-mode-p 'org-mode) (fboundp 'org-fold-show-set-visibility))
-      ;; New Org 9.6 fold-core API
-      (let ((inhibit-redisplay t)) ;; HACK: Prevent flicker due to premature redisplay
-        (org-fold-show-set-visibility 'canonical))
-    (dolist (ov (overlays-in (pos-bol) (pos-eol)))
-      (when-let (fun (overlay-get ov 'isearch-open-invisible))
-        (when (invisible-p (overlay-get ov 'invisible))
-          (funcall fun ov))))))
+  (dolist (ov (overlays-in (pos-bol) (pos-eol)))
+    (when-let (fun (overlay-get ov 'isearch-open-invisible))
+      (when (invisible-p (overlay-get ov 'invisible))
+        (funcall fun ov)))))
 
 (defun jinx--correct-highlight (overlay fun)
   "Highlight and show OVERLAY during FUN."
@@ -690,12 +687,13 @@ See `isearch-open-necessary-overlays' and `isearch-open-overlay-temporary'."
 
 (defun jinx--correct-setup ()
   "Setup minibuffer for correction."
-  (let ((message-log-max nil)
-        (inhibit-message t))
-    (use-local-map (make-composed-keymap (list jinx-correct-map) (current-local-map)))
-    (when (and (eq completing-read-function #'completing-read-default)
-               (not (bound-and-true-p vertico-mode))
-               (not (bound-and-true-p icomplete-mode)))
+  (use-local-map (make-composed-keymap (list jinx-correct-map) (current-local-map)))
+  ;; TODO Use `eager-display' on Emacs 31
+  (when (and (eq completing-read-function #'completing-read-default)
+             (not (bound-and-true-p vertico-mode))
+             (not (bound-and-true-p icomplete-mode)))
+    (let ((message-log-max nil)
+          (inhibit-message t))
       (minibuffer-completion-help))))
 
 (defun jinx--add-suggestion (list ht word group)
@@ -767,23 +765,19 @@ The word will be associated with GROUP and get a prefix key."
       (format #(" (%s)" 0 5 (face jinx-key)) (string-trim prefix))
     (get-text-property 0 'jinx--suffix cand)))
 
-(defun jinx--correct-group (word transform)
+(defun jinx--group (word transform)
   "Group WORD during completion, TRANSFORM candidate if non-nil."
   (if transform
       word
     (get-text-property 0 'jinx--group word)))
 
-(defun jinx--correct-table (suggestions)
-  "Completion table for SUGGESTIONS."
-  (lambda (str pred action)
+;; TODO: Use `completion-table-with-metadata'
+(defun jinx--table-with-metadata (table metadata)
+  "Return new completion TABLE with METADATA."
+  (lambda (string pred action)
     (if (eq action 'metadata)
-        `(metadata (category . jinx)
-                   (display-sort-function . ,#'identity)
-                   (cycle-sort-function . ,#'identity)
-                   (group-function . ,#'jinx--correct-group)
-                   (affixation-function . ,#'jinx--correct-affixation)
-                   (annotation-function . ,#'jinx--correct-annotation))
-      (complete-with-action action suggestions str pred))))
+        `(metadata . ,metadata)
+      (complete-with-action action table string pred))))
 
 (cl-defun jinx--correct-overlay (overlay &key info initial)
   "Correct word at OVERLAY.
@@ -800,8 +794,14 @@ Optionally show prompt INFO and insert INITIAL input."
                     #'jinx--correct-setup
                   (or (completing-read
                        (format "Correct ‘%s’%s: " word (or info ""))
-                       (jinx--correct-table
-                        (jinx--correct-suggestions word))
+                       (jinx--table-with-metadata
+                        (jinx--correct-suggestions word)
+                        `((category . jinx)
+                          (display-sort-function . ,#'identity)
+                          (cycle-sort-function . ,#'identity)
+                          (group-function . ,#'jinx--group)
+                          (affixation-function . ,#'jinx--correct-affixation)
+                          (annotation-function . ,#'jinx--correct-annotation)))
                        nil nil initial t word)
                       word)))))
            (len (length choice)))
@@ -856,17 +856,20 @@ Optionally show prompt INFO and insert INITIAL input."
 
 (defun jinx--load-dicts ()
   "Load dictionaries and setup syntax table."
-  (setq jinx--dicts (delq nil (mapcar #'jinx--mod-dict
-                                      (split-string jinx-languages)))
+  (setq jinx--dicts (cl-loop for lang in (split-string jinx-languages)
+                             ;; Keep a weak reference to loaded dictionaries.
+                             ;; See <gh:rrthomas/enchant#402>.
+                             for dict = (with-memoization (gethash lang jinx--dicts-hash)
+                                          (jinx--mod-dict lang))
+                             if dict collect dict)
         jinx--syntax-table (make-syntax-table jinx--base-syntax-table))
   (unless jinx--dicts
     (message "Jinx: No dictionaries available for %S" jinx-languages))
   (dolist (dict jinx--dicts)
     (cl-loop for c across (jinx--mod-wordchars dict) do
              (modify-syntax-entry c "w" jinx--syntax-table)))
-  (modify-syntax-entry ?' "w" jinx--syntax-table)
-  (modify-syntax-entry ?’ "w" jinx--syntax-table)
-  (modify-syntax-entry ?. "." jinx--syntax-table))
+  (cl-loop for (k . v) in jinx--syntax-overrides do
+           (modify-syntax-entry k v jinx--syntax-table)))
 
 (defun jinx--bounds-of-word ()
   "Return bounds of word at point using `jinx--syntax-table'."
@@ -878,6 +881,21 @@ Optionally show prompt INFO and insert INITIAL input."
         (when (re-search-forward "\\<\\w+\\>" nil t)
           (cons (match-beginning 0) (match-end 0)))))))
 
+(defun jinx--read-languages ()
+  "Read languages via `completing-read-multiple'."
+  (jinx--load-module)
+  (let ((langs (delete-dups
+                (cl-loop for (l . p) in (jinx--mod-langs) collect
+                         (propertize l 'jinx--group (format "Provider %s" p))))))
+      (string-join
+       (or (completing-read-multiple
+            (format "Change languages (%s): "
+                    (string-join (split-string jinx-languages) ", "))
+            (jinx--table-with-metadata langs `((group-function . ,#'jinx--group)))
+            nil t)
+           (user-error "No languages selected"))
+       " ")))
+
 ;;;; Save functions
 
 (defun jinx--save-personal (save key word)
@@ -926,18 +944,7 @@ by whitespace.  When called interactively, the language codes are
 read via `completing-read-multiple'.  If the prefix argument
 GLOBAL is non-nil, the languages are changed globally for all
 buffers.  See also the variable `jinx-languages'."
-  (interactive
-   (list
-    (progn
-      (jinx--load-module)
-      (string-join
-       (or (completing-read-multiple
-            (format "Change languages (%s): "
-                    (string-join (split-string jinx-languages) ", "))
-            (delete-dups (jinx--mod-langs)) nil t)
-           (user-error "No languages selected"))
-       " "))
-    current-prefix-arg))
+  (interactive (list (jinx--read-languages) current-prefix-arg))
   (unless jinx-mode (jinx-mode 1))
   (cond
    (global
@@ -1035,6 +1042,7 @@ This command dispatches to the following commands:
 
 (defun jinx-correct-select ()
   "Quick selection key for corrections."
+  (declare (completion ignore))
   (interactive)
   (let* ((keys (this-command-keys-vector))
          (word (nth (if (eq (aref keys 0) ?0)
@@ -1049,7 +1057,7 @@ This command dispatches to the following commands:
 
 (defun jinx-next (n)
   "Go to to Nth next misspelled word."
-  (interactive "p")
+  (interactive "p" jinx-mode)
   (unless (= n 0)
     (if (minibufferp)
         (throw 'jinx--goto n)
@@ -1061,7 +1069,7 @@ This command dispatches to the following commands:
 
 (defun jinx-previous (n)
   "Go to to Nth previous misspelled word."
-  (interactive "p")
+  (interactive "p" jinx-mode)
   (jinx-next (- n)))
 
 ;;;###autoload
@@ -1105,6 +1113,7 @@ This command dispatches to the following commands:
     (jit-lock-unregister #'jinx--mark-pending)
     (jinx--cleanup))))
 
+;; TODO use `:predicate' on Emacs 29
 (defcustom global-jinx-modes '(text-mode prog-mode conf-mode)
   "List of modes where Jinx should be enabled.
 The variable can either be t, nil or a list of t, nil, mode
@@ -1122,7 +1131,7 @@ symbols or elements of the form (not modes)."
                       buffer-read-only
                       (buffer-base-buffer) ;; Do not enable in indirect buffers
                       (eq (aref (buffer-name) 0) ?\s)))
-             ;; TODO backport `easy-mmode--globalized-predicate-p'
+             ;; TODO use `:predicate' on Emacs 29
              (or (eq t global-jinx-modes)
                  (eq t (cl-loop for p in global-jinx-modes thereis
                                 (pcase-exhaustive p
@@ -1132,9 +1141,5 @@ symbols or elements of the form (not modes)."
                                   (`(not . ,m) (and (seq-some #'derived-mode-p m) 0)))))))
     (jinx-mode 1)))
 
-(put #'jinx-correct-select 'completion-predicate #'ignore)
-(put #'jinx-next 'command-modes '(jinx-mode))
-(put #'jinx-previous 'command-modes '(jinx-mode))
-
 (provide 'jinx)
 ;;; jinx.el ends here
